@@ -14,7 +14,7 @@ from typing import (
 import kitty.fast_data_types as fast_data_types
 
 from .constants import is_macos, shell_path, terminfo_dir
-from .options_stub import Options
+from .types import run_once
 
 try:
     from typing import TypedDict
@@ -24,8 +24,9 @@ except ImportError:
 
 if is_macos:
     from kitty.fast_data_types import (
-            cmdline_of_process, cwd_of_process as _cwd, environ_of_process as _environ_of_process,
-            process_group_map as _process_group_map
+        cmdline_of_process, cwd_of_process as _cwd,
+        environ_of_process as _environ_of_process,
+        process_group_map as _process_group_map
     )
 
     def cwd_of_process(pid: int) -> str:
@@ -40,13 +41,16 @@ if is_macos:
 else:
 
     def cmdline_of_process(pid: int) -> List[str]:
-        return list( [ 'kitty' ] )
+        with open('/proc/{}/cmdline'.format(pid), 'rb') as f:
+            return list(filter(None, f.read().decode('utf-8').split('\0')))
 
     def cwd_of_process(pid: int) -> str:
-        return "/usr/local/bin"
+        ans = '/proc/{}/cwd'.format(pid)
+        return os.path.realpath(ans)
 
     def _environ_of_process(pid: int) -> str:
-        return ""
+        with open('/proc/{}/environ'.format(pid), 'rb') as f:
+            return f.read().decode('utf-8')
 
     def process_group_map() -> DefaultDict[int, List[int]]:
         ans: DefaultDict[int, List[int]] = defaultdict(list)
@@ -168,6 +172,11 @@ def openpty() -> Tuple[int, int]:
     return master, slave
 
 
+@run_once
+def getpid() -> str:
+    return str(os.getpid())
+
+
 class ProcessDesc(TypedDict):
     cwd: Optional[str]
     pid: int
@@ -184,7 +193,6 @@ class Child:
         self,
         argv: Sequence[str],
         cwd: str,
-        opts: Options,
         stdin: Optional[bytes] = None,
         env: Optional[Dict[str, str]] = None,
         cwd_from: Optional[int] = None,
@@ -201,7 +209,6 @@ class Child:
         else:
             cwd = os.path.expandvars(os.path.expanduser(cwd or os.getcwd()))
         self.cwd = os.path.abspath(cwd)
-        self.opts = opts
         self.stdin = stdin
         self.env = env or {}
 
@@ -214,8 +221,9 @@ class Child:
                     'lc_ctype_before_python') and not getattr(default_env, 'lc_ctype_set_by_user', False):
                 del env['LC_CTYPE']
             env.update(self.env)
-            env['TERM'] = self.opts.term
+            env['TERM'] = fast_data_types.get_options().term
             env['COLORTERM'] = 'truecolor'
+            env['KITTY_PID'] = getpid()
             if self.cwd:
                 # needed in case cwd is a symlink, in which case shells
                 # can use it to display the current directory name rather
