@@ -2,6 +2,8 @@
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import os
+import stat
 import weakref
 from collections import deque
 from contextlib import suppress
@@ -15,16 +17,16 @@ from typing import (
 from .borders import Borders
 from .child import Child
 from .cli_stub import CLIOptions
-from .constants import appname
+from .constants import appname, kitty_exe
 from .fast_data_types import (
-    add_tab, attach_window, detach_window, get_boss, mark_tab_bar_dirty,
-    next_window_id, remove_tab, remove_window, ring_bell, set_active_tab,
-    set_active_window, swap_tabs, sync_os_window_title
+    add_tab, attach_window, detach_window, get_boss, get_options,
+    mark_tab_bar_dirty, next_window_id, remove_tab, remove_window, ring_bell,
+    set_active_tab, set_active_window, swap_tabs, sync_os_window_title
 )
 from .layout.base import Layout, Rect
 from .layout.interface import create_layout_object_for, evict_cached_layouts
-from .options_stub import Options
 from .tab_bar import TabBar, TabBarData
+from .types import ac
 from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
 from .utils import log_error, platform_window_id, resolved_shell
 from .window import Watchers, Window, WindowDict
@@ -36,6 +38,7 @@ class TabDict(TypedDict):
     is_focused: bool
     title: str
     layout: str
+    layout_state: Dict[str, Any]
     windows: List[WindowDict]
     active_window_history: List[int]
 
@@ -48,6 +51,7 @@ class SpecialWindowInstance(NamedTuple):
     cwd: Optional[str]
     overlay_for: Optional[int]
     env: Optional[Dict[str, str]]
+    watchers: Optional[Watchers]
 
 
 def SpecialWindow(
@@ -57,9 +61,10 @@ def SpecialWindow(
     cwd_from: Optional[int] = None,
     cwd: Optional[str] = None,
     overlay_for: Optional[int] = None,
-    env: Optional[Dict[str, str]] = None
+    env: Optional[Dict[str, str]] = None,
+    watchers: Optional[Watchers] = None
 ) -> SpecialWindowInstance:
-    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env)
+    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers)
 
 
 def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64) -> None:
@@ -85,10 +90,10 @@ class Tab:  # {{{
         self.id: int = add_tab(self.os_window_id)
         if not self.id:
             raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
-        self.opts, self.args = tab_manager.opts, tab_manager.args
+        self.args = tab_manager.args
         self.name = getattr(session_tab, 'name', '')
-        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or self.opts.enabled_layouts]
-        self.borders = Borders(self.os_window_id, self.id, self.opts)
+        self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
+        self.borders = Borders(self.os_window_id, self.id)
         self.windows = WindowList(self)
         for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
             setattr(self, which + '_window', partial(self.nth_window, num=i))
@@ -111,6 +116,14 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
+    def apply_options(self) -> None:
+        for window in self:
+            window.apply_options()
+        self.enabled_layouts = [x.lower() for x in get_options().enabled_layouts] or ['tall']
+        if self.current_layout.name not in self.enabled_layouts:
+            self._set_current_layout(self.enabled_layouts[0])
+        self.relayout()
+
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
@@ -135,10 +148,11 @@ class Tab:  # {{{
 
     def startup(self, session_tab: 'SessionTab') -> None:
         for cmd in session_tab.windows:
-            if isinstance(cmd, (SpecialWindowInstance,)):
+            if isinstance(cmd, SpecialWindowInstance):
                 self.new_special_window(cmd)
             else:
-                self.new_window(cmd=cmd)
+                from .launch import launch
+                launch(get_boss(), cmd.opts, cmd.args, target_tab=self, force_target_tab=True)
         self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -209,6 +223,7 @@ class Tab:  # {{{
     def create_layout_object(self, name: str) -> Layout:
         return create_layout_object_for(name, self.os_window_id, self.id)
 
+    @ac('lay', 'Go to the next enabled layout')
     def next_layout(self) -> None:
         if len(self.enabled_layouts) > 1:
             for i, layout_name in enumerate(self.enabled_layouts):
@@ -221,11 +236,19 @@ class Tab:  # {{{
             self._set_current_layout(nl)
             self.relayout()
 
+    @ac('lay', 'Go to the previously used layout')
     def last_used_layout(self) -> None:
         if len(self.enabled_layouts) > 1 and self._last_used_layout and self._last_used_layout != self._current_layout_name:
             self._set_current_layout(self._last_used_layout)
             self.relayout()
 
+    @ac('lay', '''
+        Switch to the named layout
+
+        For example::
+
+            map f1 goto_layout tall
+        ''')
     def goto_layout(self, layout_name: str, raise_exception: bool = False) -> None:
         layout_name = layout_name.lower()
         if layout_name not in self.enabled_layouts:
@@ -236,6 +259,21 @@ class Tab:  # {{{
         self._set_current_layout(layout_name)
         self.relayout()
 
+    @ac('lay', '''
+        Toggle the named layout
+
+        Switches to the named layout if another layout is current, otherwise
+        switches to the last used layout. Useful to "zoom" a window temporarily
+        by switching to the stack layout. For example::
+
+            map f1 toggle_layout stack
+        ''')
+    def toggle_layout(self, layout_name: str) -> None:
+        if self._current_layout_name == layout_name:
+            self.last_used_layout()
+        else:
+            self.goto_layout(layout_name)
+
     def resize_window_by(self, window_id: int, increment: float, is_horizontal: bool) -> Optional[str]:
         increment_as_percent = self.current_layout.bias_increment_for_cell(is_horizontal) * increment
         if self.current_layout.modify_size_of_window(self.windows, window_id, increment_as_percent, is_horizontal):
@@ -243,6 +281,11 @@ class Tab:  # {{{
             return None
         return 'Could not resize'
 
+    @ac('win', '''
+        Resize the active window by the specified amount
+
+        See :ref:`window_resizing` for details.
+        ''')
     def resize_window(self, quality: str, increment: int) -> None:
         if increment < 1:
             raise ValueError(increment)
@@ -251,12 +294,15 @@ class Tab:  # {{{
         w = self.active_window
         if w is not None and self.resize_window_by(
                 w.id, increment, is_horizontal) is not None:
-            ring_bell()
+            if get_options().enable_audio_bell:
+                ring_bell()
 
+    @ac('win', 'Reset window sizes undoing any dynamic resizing of windows')
     def reset_window_sizes(self) -> None:
         if self.current_layout.remove_all_biases():
             self.relayout()
 
+    @ac('lay', 'Perform a layout specific action. See :doc:`layouts` for details')
     def layout_action(self, action_name: str, args: Sequence[str]) -> None:
         ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
@@ -274,11 +320,47 @@ class Tab:  # {{{
         env: Optional[Dict[str, str]] = None,
         allow_remote_control: bool = False
     ) -> Child:
+        check_for_suitability = True
         if cmd is None:
             if use_shell:
-                cmd = resolved_shell(self.opts)
+                cmd = resolved_shell(get_options())
+                check_for_suitability = False
             else:
-                cmd = self.args.args or resolved_shell(self.opts)
+                if self.args.args:
+                    cmd = list(self.args.args)
+                else:
+                    cmd = resolved_shell(get_options())
+                    check_for_suitability = False
+                cmd = self.args.args or resolved_shell(get_options())
+        if check_for_suitability:
+            old_exe = cmd[0]
+            if not os.path.isabs(old_exe):
+                import shutil
+                actual_exe = shutil.which(old_exe)
+                old_exe = actual_exe if actual_exe else os.path.abspath(old_exe)
+            try:
+                is_executable = os.access(old_exe, os.X_OK)
+            except OSError:
+                pass
+            else:
+                try:
+                    st = os.stat(old_exe)
+                except OSError:
+                    pass
+                else:
+                    if stat.S_ISDIR(st.st_mode):
+                        cwd = old_exe
+                        cmd = resolved_shell(get_options())
+                    elif not is_executable:
+                        import shlex
+                        with suppress(OSError):
+                            with open(old_exe) as f:
+                                cmd = [kitty_exe(), '+hold']
+                                if f.read(2) == '#!':
+                                    line = f.read(4096).splitlines()[0]
+                                    cmd += shlex.split(line) + [old_exe]
+                                else:
+                                    cmd += [resolved_shell(get_options())[0], cmd[0]]
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
@@ -286,7 +368,7 @@ class Tab:  # {{{
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, self.opts, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
+        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
         ans.fork()
         return ans
 
@@ -314,7 +396,7 @@ class Tab:  # {{{
         child = self.launch_child(
             use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env, allow_remote_control=allow_remote_control)
         window = Window(
-            self, child, self.opts, self.args, override_title=override_title,
+            self, child, self.args, override_title=override_title,
             copy_colors_from=copy_colors_from, watchers=watchers
         )
         # Must add child before laying out so that resize_pty succeeds
@@ -333,21 +415,23 @@ class Tab:  # {{{
             special_window: SpecialWindowInstance,
             location: Optional[str] = None,
             copy_colors_from: Optional[Window] = None,
-            allow_remote_control: bool = False
+            allow_remote_control: bool = False,
     ) -> Window:
         return self.new_window(
             use_shell=False, cmd=special_window.cmd, stdin=special_window.stdin,
             override_title=special_window.override_title,
             cwd_from=special_window.cwd_from, cwd=special_window.cwd, overlay_for=special_window.overlay_for,
             env=special_window.env, location=location, copy_colors_from=copy_colors_from,
-            allow_remote_control=allow_remote_control
+            allow_remote_control=allow_remote_control, watchers=special_window.watchers
         )
 
+    @ac('win', 'Close the currently active window')
     def close_window(self) -> None:
         w = self.active_window
         if w is not None:
             self.remove_window(w)
 
+    @ac('win', 'Close all windows in the tab other than the currently active window')
     def close_other_windows_in_tab(self) -> None:
         if len(self.windows) > 1:
             active_window = self.active_window
@@ -386,6 +470,13 @@ class Tab:  # {{{
         if self.windows:
             return self.current_layout.nth_window(self.windows, n)
 
+    @ac('win', '''
+        Focus the nth window if positive or the previously active windows if negative
+
+        For example, to ficus the previously active window::
+
+            map ctrl+p nth_window -1
+        ''')
     def nth_window(self, num: int = 0) -> None:
         if self.windows:
             if num < 0:
@@ -399,9 +490,11 @@ class Tab:  # {{{
             self.current_layout.next_window(self.windows, delta)
             self.relayout_borders()
 
+    @ac('win', 'Focus the next window in the current tab')
     def next_window(self) -> None:
         self._next_window()
 
+    @ac('win', 'Focus the previous window in the current tab')
     def previous_window(self) -> None:
         self._next_window(-1)
 
@@ -424,11 +517,27 @@ class Tab:  # {{{
         if candidates:
             return self.most_recent_group(candidates)
 
+    @ac('win', '''
+        Focus the neighboring window in the current tab
+
+        For example::
+
+            map ctrl+left neighboring_window left
+            map ctrl+down neighboring_window bottom
+        ''')
     def neighboring_window(self, which: EdgeLiteral) -> None:
         neighbor = self.neighboring_group_id(which)
         if neighbor:
             self.windows.set_active_group(neighbor)
 
+    @ac('win', '''
+        Move the window in the specified direction
+
+        For example::
+
+            map ctrl+left move_window left
+            map ctrl+down move_window bottom
+        ''')
     def move_window(self, delta: Union[EdgeLiteral, int] = 1) -> None:
         if isinstance(delta, int):
             if self.current_layout.move_window(self.windows, delta):
@@ -439,20 +548,23 @@ class Tab:  # {{{
                 if self.current_layout.move_window_to_group(self.windows, neighbor):
                     self.relayout()
 
+    @ac('win', 'Move active window to the top (make it the first window)')
     def move_window_to_top(self) -> None:
-        n = self.windows.num_groups
-        if n > 1:
-            self.move_window(1 - n)
+        n = self.windows.active_group_idx
+        if n > 0:
+            self.move_window(-n)
 
+    @ac('win', 'Move active window forward (swap it with the next window)')
     def move_window_forward(self) -> None:
         self.move_window()
 
+    @ac('win', 'Move active window backward (swap it with the previous window)')
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, active_window: Optional[Window]) -> Generator[WindowDict, None, None]:
+    def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
         for w in self:
-            yield w.as_dict(is_focused=w is active_window)
+            yield w.as_dict(is_focused=w is active_window, is_self=w is self_window)
 
     def matches(self, field: str, pat: Pattern) -> bool:
         if field == 'id':
@@ -467,6 +579,10 @@ class Tab:  # {{{
     def __len__(self) -> int:
         return len(self.windows)
 
+    @property
+    def num_window_groups(self) -> int:
+        return self.windows.num_groups
+
     def __contains__(self, window: Window) -> bool:
         return window in self.windows
 
@@ -488,14 +604,16 @@ class Tab:  # {{{
 
 class TabManager:  # {{{
 
-    def __init__(self, os_window_id: int, opts: Options, args: CLIOptions, startup_session: Optional[SessionType] = None):
+    def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
+        self.wm_class = wm_class
+        self.wm_name = wm_name
         self.last_active_tab_id = None
-        self.opts, self.args = opts, args
-        self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
+        self.args = args
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
         self.tabs: List[Tab] = []
         self.active_tab_history: Deque[int] = deque()
-        self.tab_bar = TabBar(self.os_window_id, opts)
+        self.tab_bar = TabBar(self.os_window_id)
         self._active_tab_idx = 0
 
         if startup_session is not None:
@@ -509,6 +627,9 @@ class TabManager:  # {{{
 
     @active_tab_idx.setter
     def active_tab_idx(self, val: int) -> None:
+        new_active_tab_idx = max(0, min(val, len(self.tabs) - 1))
+        if new_active_tab_idx == self._active_tab_idx:
+            return
         try:
             old_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
         except Exception:
@@ -516,7 +637,7 @@ class TabManager:  # {{{
         else:
             assert old_active_tab is not None
             add_active_id_to_history(self.active_tab_history, old_active_tab.id)
-        self._active_tab_idx = max(0, min(val, len(self.tabs) - 1))
+        self._active_tab_idx = new_active_tab_idx
         try:
             new_active_tab: Optional[Tab] = self.tabs[self._active_tab_idx]
         except Exception:
@@ -537,7 +658,7 @@ class TabManager:  # {{{
 
     @property
     def tab_bar_should_be_visible(self) -> bool:
-        return len(self.tabs) >= self.opts.tab_bar_min_tabs
+        return len(self.tabs) >= get_options().tab_bar_min_tabs
 
     def _add_tab(self, tab: Tab) -> None:
         visible_before = self.tab_bar_should_be_visible
@@ -600,6 +721,18 @@ class TabManager:  # {{{
         if len(self.tabs) > 1:
             self.set_active_tab_idx((self.active_tab_idx + len(self.tabs) + delta) % len(self.tabs))
 
+    def tab_at_location(self, loc: str) -> Optional[Tab]:
+        if loc == 'prev':
+            if self.active_tab_history:
+                old_active_tab_id = self.active_tab_history[-1]
+                for idx, tab in enumerate(self.tabs):
+                    if tab.id == old_active_tab_id:
+                        return tab
+        elif loc in ('left', 'right'):
+            delta = -1 if loc == 'left' else 1
+            idx = (len(self.tabs) + self.active_tab_idx + delta) % len(self.tabs)
+            return self.tabs[idx]
+
     def goto_tab(self, tab_num: int) -> None:
         if tab_num >= len(self.tabs):
             tab_num = max(0, len(self.tabs) - 1)
@@ -621,14 +754,15 @@ class TabManager:  # {{{
     def __len__(self) -> int:
         return len(self.tabs)
 
-    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window]) -> Generator[TabDict, None, None]:
+    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
         for tab in self:
             yield {
                 'id': tab.id,
                 'is_focused': tab is active_tab,
                 'title': tab.name or tab.title,
                 'layout': str(tab.current_layout.name),
-                'windows': list(tab.list_windows(active_window)),
+                'layout_state': tab.current_layout.layout_state(),
+                'windows': list(tab.list_windows(active_window, self_window)),
                 'active_window_history': list(tab.windows.active_window_history),
             }
 
@@ -642,7 +776,10 @@ class TabManager:  # {{{
 
     @property
     def active_tab(self) -> Optional[Tab]:
-        return self.tabs[self.active_tab_idx] if self.tabs else None
+        try:
+            return self.tabs[self.active_tab_idx] if self.tabs else None
+        except Exception:
+            return None
 
     @property
     def active_window(self) -> Optional[Window]:
@@ -705,28 +842,41 @@ class TabManager:  # {{{
         return self.tabs[idx]
 
     def remove(self, tab: Tab) -> None:
+        active_tab_before_removal = self.active_tab
         self._remove_tab(tab)
-        next_active_tab = -1
+        active_tab = self.active_tab
+        active_tab_needs_to_change = (active_tab is None and (active_tab_before_removal is None or active_tab_before_removal is tab)) or active_tab is tab
         while True:
             try:
                 self.active_tab_history.remove(tab.id)
             except ValueError:
                 break
 
-        if self.opts.tab_switch_strategy == 'previous':
-            while self.active_tab_history and next_active_tab < 0:
-                tab_id = self.active_tab_history.pop()
-                for idx, qtab in enumerate(self.tabs):
-                    if qtab.id == tab_id:
-                        next_active_tab = idx
-                        break
-        elif self.opts.tab_switch_strategy == 'left':
-            next_active_tab = max(0, self.active_tab_idx - 1)
+        if active_tab_needs_to_change:
+            next_active_tab = -1
+            if get_options().tab_switch_strategy == 'previous':
+                while self.active_tab_history and next_active_tab < 0:
+                    tab_id = self.active_tab_history.pop()
+                    for idx, qtab in enumerate(self.tabs):
+                        if qtab.id == tab_id:
+                            next_active_tab = idx
+                            break
+            elif get_options().tab_switch_strategy == 'left':
+                next_active_tab = max(0, self.active_tab_idx - 1)
+            elif get_options().tab_switch_strategy == 'right':
+                next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
 
-        if next_active_tab < 0:
-            next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
+            if next_active_tab < 0:
+                next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
 
-        self._set_active_tab(next_active_tab)
+            self._set_active_tab(next_active_tab)
+        elif active_tab_before_removal is not None:
+            try:
+                idx = self.tabs.index(active_tab_before_removal)
+            except Exception:
+                pass
+            else:
+                self._active_tab_idx = idx
         self.mark_tab_bar_dirty()
         tab.destroy()
 
@@ -737,19 +887,25 @@ class TabManager:  # {{{
         for t in self.tabs:
             title = (t.name or t.title or appname).strip()
             needs_attention = False
+            has_activity_since_last_focus = False
             for w in t:
                 if w.needs_attention:
                     needs_attention = True
-                    break
+                if w.has_activity_since_last_focus:
+                    has_activity_since_last_focus = True
             ans.append(TabBarData(
                 title, t is at, needs_attention,
-                len(t), t.current_layout.name or ''
+                len(t), t.num_window_groups, t.current_layout.name or '',
+                has_activity_since_last_focus
             ))
         return ans
 
-    def activate_tab_at(self, x: int) -> None:
+    def activate_tab_at(self, x: int, is_double: bool = False) -> None:
         i = self.tab_bar.tab_at(x)
-        if i is not None:
+        if i is None:
+            if is_double:
+                self.new_tab()
+        else:
             self.set_active_tab_idx(i)
 
     @property
@@ -762,4 +918,13 @@ class TabManager:  # {{{
         self.tab_bar.destroy()
         del self.tab_bar
         del self.tabs
+
+    def apply_options(self) -> None:
+        for tab in self:
+            tab.apply_options()
+        self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
+        self.tab_bar.apply_options()
+        self.update_tab_bar_data()
+        self.mark_tab_bar_dirty()
+        self.tab_bar.layout()
 # }}}
