@@ -8,36 +8,41 @@ import sys
 import weakref
 from collections import deque
 from enum import IntEnum
+from functools import partial
+from gettext import gettext as _
 from itertools import chain
 from typing import (
-    Any, Callable, Deque, Dict, Iterable, List, Optional, Pattern, Sequence,
-    Tuple, Union
+    Any, Callable, Deque, Dict, Iterable, List, NamedTuple, Optional, Pattern,
+    Sequence, Tuple, Union
 )
 
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
 from .config import build_ansi_color_table
-from .constants import ScreenGeometry, WindowGeometry, appname, wakeup
+from .constants import appname, is_macos, wakeup
 from .fast_data_types import (
     BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
-    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CSI, DCS, DECORATION, DIM,
+    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK,
+    CURSOR_UNDERLINE, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
     GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
-    MARK, MARK_MASK, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE,
-    STRIKETHROUGH, TINT_PROGRAM, Screen, add_window, cell_size_for_window,
-    compile_program, get_boss, get_clipboard_string, init_cell_program,
-    pt_to_px, set_clipboard_string, set_titlebar_color, set_window_padding,
-    set_window_render_data, update_window_title, update_window_visibility,
-    viewport_for_window
+    MARK, MARK_MASK, NO_CURSOR_SHAPE, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE,
+    SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, KeyEvent, Screen, add_timer,
+    add_window, cell_size_for_window, click_mouse_url, compile_program,
+    encode_key_for_tty, get_boss, get_clipboard_string, get_options,
+    init_cell_program, mouse_selection, pt_to_px, set_clipboard_string,
+    set_titlebar_color, set_window_padding, set_window_render_data,
+    update_window_title, update_window_visibility, viewport_for_window
 )
-from .keys import defines, extended_key_event, keyboard_mode_name
-from .options_stub import Options
+from .keys import keyboard_mode_name, mod_mask
+from .notify import NotificationCommand, handle_notification_cmd
+from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
+from .types import MouseEvent, ScreenGeometry, WindowGeometry, ac
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    color_as_int, get_primary_selection, load_shaders, open_cmd, open_url,
-    parse_color_set, read_shell_environment, sanitize_title,
-    set_primary_selection
+    color_as_int, get_primary_selection, load_shaders, log_error, open_cmd,
+    open_url, parse_color_set, sanitize_title, set_primary_selection
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
@@ -52,6 +57,9 @@ class WindowDict(TypedDict):
     cmdline: List[str]
     env: Dict[str, str]
     foreground_processes: List[ProcessDesc]
+    is_self: bool
+    lines: int
+    columns: int
 
 
 class PipeData(TypedDict):
@@ -64,6 +72,12 @@ class PipeData(TypedDict):
     text: str
 
 
+class ClipboardPending(NamedTuple):
+    where: str
+    data: str
+    truncated: bool = False
+
+
 class DynamicColor(IntEnum):
     default_fg, default_bg, cursor_color, highlight_fg, highlight_bg = range(1, 6)
 
@@ -86,9 +100,48 @@ class Watcher:
 
 class Watchers:
 
+    on_resize: List[Watcher]
+    on_close: List[Watcher]
+    on_focus_change: List[Watcher]
+
     def __init__(self) -> None:
-        self.on_resize: List[Watcher] = []
-        self.on_close: List[Watcher] = []
+        self.on_resize = []
+        self.on_close = []
+        self.on_focus_change = []
+
+    def add(self, others: 'Watchers') -> None:
+        def merge(base: List[Watcher], other: List[Watcher]) -> None:
+            for x in other:
+                if x not in base:
+                    base.append(x)
+        merge(self.on_resize, others.on_resize)
+        merge(self.on_close, others.on_close)
+        merge(self.on_focus_change, others.on_focus_change)
+
+    def clear(self) -> None:
+        del self.on_close[:], self.on_resize[:], self.on_focus_change[:]
+
+    def copy(self) -> 'Watchers':
+        ans = Watchers()
+        ans.on_close = self.on_close[:]
+        ans.on_resize = self.on_resize[:]
+        ans.on_focus_change = self.on_focus_change
+        return ans
+
+    @property
+    def has_watchers(self) -> bool:
+        return bool(self.on_close or self.on_resize or self.on_focus_change)
+
+
+def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data: Dict[str, Any]) -> None:
+
+    def callback(timer_id: Optional[int]) -> None:
+        w = windowref()
+        if w is not None:
+            watchers: List[Watcher] = getattr(w.watchers, which)
+            w.call_watchers(watchers, data)
+
+    add_timer(callback, 0, False)
 
 
 def calculate_gl_geometry(window_geometry: WindowGeometry, viewport_width: int, viewport_height: int, cell_width: int, cell_height: int) -> ScreenGeometry:
@@ -100,6 +153,58 @@ def calculate_gl_geometry(window_geometry: WindowGeometry, viewport_width: int,
     return ScreenGeometry(xstart, ystart, window_geometry.xnum, window_geometry.ynum, dx, dy)
 
 
+def as_text(
+    screen: Screen,
+    as_ansi: bool = False,
+    add_history: bool = False,
+    add_wrap_markers: bool = False,
+    alternate_screen: bool = False,
+    add_cursor: bool = False
+) -> str:
+    lines: List[str] = []
+    add_history = add_history and not (screen.is_using_alternate_linebuf() ^ alternate_screen)
+    if alternate_screen:
+        f = screen.as_text_alternate
+    else:
+        f = screen.as_text_non_visual if add_history else screen.as_text
+    f(lines.append, as_ansi, add_wrap_markers)
+    ctext = ''
+    if add_cursor:
+        ctext += '\x1b[?25' + ('h' if screen.cursor_visible else 'l')
+        ctext += f'\x1b[{screen.cursor.y + 1};{screen.cursor.x + 1}H'
+        shape = screen.cursor.shape
+        if shape == NO_CURSOR_SHAPE:
+            ctext += '\x1b[?12' + ('h' if screen.cursor.blink else 'l')
+        else:
+            code = {CURSOR_BLOCK: 1, CURSOR_UNDERLINE: 3, CURSOR_BEAM: 5}[shape]
+            if not screen.cursor.blink:
+                code += 1
+            ctext += f'\x1b[{code} q'
+
+    if add_history:
+        h: List[str] = []
+        pht = screen.historybuf.pagerhist_as_text()
+        if pht:
+            h.append(pht)
+        if h and (not as_ansi or not add_wrap_markers):
+            sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
+            h = list(map(sanitizer, h))
+        screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
+        if h:
+            if not screen.linebuf.is_continued(0):
+                h[-1] += '\n'
+            if as_ansi:
+                h[-1] += '\x1b[m'
+        ans = ''.join(chain(h, lines))
+        if ctext:
+            ans += ctext
+        return ans
+    ans = ''.join(lines)
+    if ctext:
+        ans += ctext
+    return ans
+
+
 class LoadShaderPrograms:
 
     use_selection_fg = True
@@ -167,7 +272,7 @@ def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str
     pat = getattr(text_sanitizer, 'pat', None)
     if pat is None:
         import re
-        pat = re.compile(r'\033\[.+?m')
+        pat = re.compile('\033\\[.*?m')
         setattr(text_sanitizer, 'pat', pat)
 
     ansi, wrap_markers = not as_ansi, not add_wrap_markers
@@ -211,13 +316,14 @@ class Window:
         self,
         tab: TabType,
         child: ChildType,
-        opts: Options,
         args: CLIOptions,
         override_title: Optional[str] = None,
         copy_colors_from: Optional['Window'] = None,
         watchers: Optional[Watchers] = None
     ):
         self.watchers = watchers or Watchers()
+        self.current_mouse_event_button = 0
+        self.prev_osc99_cmd = NotificationCommand()
         self.action_on_close: Optional[Callable] = None
         self.action_on_removal: Optional[Callable] = None
         self.current_marker_spec: Optional[Tuple[str, Union[str, Tuple[Tuple[int, str], ...]]]] = None
@@ -237,13 +343,14 @@ class Window:
         self.tab_id = tab.id
         self.os_window_id = tab.os_window_id
         self.tabref: Callable[[], Optional[TabType]] = weakref.ref(tab)
-        self.clipboard_control_buffers = {'p': '', 'c': ''}
+        self.clipboard_pending: Optional[ClipboardPending] = None
         self.destroyed = False
         self.geometry: WindowGeometry = WindowGeometry(0, 0, 0, 0, 0, 0)
         self.needs_layout = True
         self.is_visible_in_layout: bool = True
-        self.child, self.opts = child, opts
+        self.child = child
         cell_width, cell_height = cell_size_for_window(self.os_window_id)
+        opts = get_options()
         self.screen: Screen = Screen(self, 24, 80, opts.scrollback_lines, cell_width, cell_height, self.id)
         if copy_colors_from is not None:
             self.screen.copy_colors_from(copy_colors_from.screen)
@@ -262,18 +369,19 @@ class Window:
         q = getattr(self.margin, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
+        opts = get_options()
         if is_single_window:
-            q = getattr(self.opts.single_window_margin_width, edge)
+            q = getattr(opts.single_window_margin_width, edge)
             if q > -0.1:
                 return pt_to_px(q, self.os_window_id)
-        q = getattr(self.opts.window_margin_width, edge)
+        q = getattr(opts.window_margin_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def effective_padding(self, edge: EdgeLiteral) -> int:
         q = getattr(self.padding, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
-        q = getattr(self.opts.window_padding_width, edge)
+        q = getattr(get_options().window_padding_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def update_effective_padding(self) -> None:
@@ -289,7 +397,18 @@ class Window:
             self.update_effective_padding()
 
     def effective_border(self) -> int:
-        return pt_to_px(self.opts.window_border_width, self.os_window_id)
+        val, unit = get_options().window_border_width
+        if unit == 'pt':
+            val = max(1 if val > 0 else 0, pt_to_px(val, self.os_window_id))
+        else:
+            val = round(val)
+        return int(val)
+
+    def apply_options(self) -> None:
+        opts = get_options()
+        self.update_effective_padding()
+        self.change_titlebar_color()
+        setup_colors(self.screen, opts)
 
     @property
     def title(self) -> str:
@@ -299,7 +418,7 @@ class Window:
         return 'Window(title={}, id={})'.format(
                 self.title, self.id)
 
-    def as_dict(self, is_focused: bool = False) -> WindowDict:
+    def as_dict(self, is_focused: bool = False, is_self: bool = False) -> WindowDict:
         return dict(
             id=self.id,
             is_focused=is_focused,
@@ -308,7 +427,10 @@ class Window:
             cwd=self.child.current_cwd or self.child.cwd,
             cmdline=self.child.cmdline,
             env=self.child.environ,
-            foreground_processes=self.child.foreground_processes
+            foreground_processes=self.child.foreground_processes,
+            is_self=is_self,
+            lines=self.screen.lines,
+            columns=self.screen.columns,
         )
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -344,7 +466,7 @@ class Window:
             return False
         assert not isinstance(pat, tuple)
 
-        if field == 'id':
+        if field in ('id', 'window_id'):
             return True if pat.pattern == str(self.id) else False
         if field == 'pid':
             return True if pat.pattern == str(self.child.pid) else False
@@ -383,7 +505,7 @@ class Window:
             self.screen.resize(new_geometry.ynum, new_geometry.xnum)
             sg = self.update_position(new_geometry)
             self.needs_layout = False
-            self.call_watchers(self.watchers.on_resize, {'old_geometry': self.geometry, 'new_geometry': new_geometry})
+            call_watchers(weakref.ref(self), 'on_resize', {'old_geometry': self.geometry, 'new_geometry': new_geometry})
         else:
             sg = self.update_position(new_geometry)
         current_pty_size = (
@@ -407,6 +529,11 @@ class Window:
     def close(self) -> None:
         get_boss().close_window(self)
 
+    @ac('misc', '''
+        Send the specified text to the active window
+
+        For details, see :sc:`send_text`.
+        ''')
     def send_text(self, *args: str) -> bool:
         mode = keyboard_mode_name(self.screen)
         required_mode_, text = args[-2:]
@@ -420,7 +547,7 @@ class Window:
     def write_to_child(self, data: Union[str, bytes]) -> None:
         if data:
             if get_boss().child_monitor.needs_write(self.id, data) is not True:
-                print('Failed to write to child %d as it does not exist' % self.id, file=sys.stderr)
+                log_error(f'Failed to write to child {self.id} as it does not exist')
 
     def title_updated(self) -> None:
         update_window_title(self.os_window_id, self.tab_id, self.id, self.title)
@@ -434,20 +561,90 @@ class Window:
         self.override_title = title or None
         self.title_updated()
 
+    def desktop_notify(self, osc_code: int, raw_data: str) -> None:
+        cmd = handle_notification_cmd(osc_code, raw_data, self.id, self.prev_osc99_cmd)
+        if cmd is not None and osc_code == 99:
+            self.prev_osc99_cmd = cmd
+
     # screen callbacks {{{
     def use_utf8(self, on: bool) -> None:
         get_boss().child_monitor.set_iutf8_winid(self.id, on)
 
+    def on_mouse_event(self, event: Dict[str, Any]) -> bool:
+        event['mods'] = event.get('mods', 0) & mod_mask
+        ev = MouseEvent(**event)
+        self.current_mouse_event_button = ev.button
+        action = get_options().mousemap.get(ev)
+        if action is None:
+            return False
+        return get_boss().dispatch_action(action, window_for_dispatch=self, dispatch_type='MouseEvent')
+
+    def open_url(self, url: str, hyperlink_id: int, cwd: Optional[str] = None) -> None:
+        opts = get_options()
+        if hyperlink_id:
+            if not opts.allow_hyperlinks:
+                return
+            from urllib.parse import unquote, urlparse, urlunparse
+            try:
+                purl = urlparse(url)
+            except Exception:
+                return
+            if (not purl.scheme or purl.scheme == 'file'):
+                if purl.netloc:
+                    from socket import gethostname
+                    try:
+                        hostname = gethostname()
+                    except Exception:
+                        hostname = ''
+                    remote_hostname = purl.netloc.partition(':')[0]
+                    if remote_hostname and remote_hostname != hostname and remote_hostname != 'localhost':
+                        self.handle_remote_file(purl.netloc, unquote(purl.path))
+                        return
+                    url = urlunparse(purl._replace(netloc=''))
+            if opts.allow_hyperlinks & 0b10:
+                from kittens.tui.operations import styled
+                get_boss()._run_kitten('ask', ['--type=choices', '--message', _(
+                    'What would you like to do with this URL:\n') +
+                    styled(unquote(url), fg='yellow'),
+                    '--choice=o:Open', '--choice=c:Copy to clipboard', '--choice=n;red:Nothing'
+                    ],
+                    window=self,
+                    custom_callback=partial(self.hyperlink_open_confirmed, url, cwd)
+                )
+                return
+        get_boss().open_url(url, cwd=cwd)
+
+    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], data: Dict[str, Any], *a: Any) -> None:
+        q = data['response']
+        if q == 'o':
+            get_boss().open_url(url, cwd=cwd)
+        elif q == 'c':
+            set_clipboard_string(url)
+
+    def handle_remote_file(self, netloc: str, remote_path: str) -> None:
+        from kittens.ssh.main import get_connection_data
+        args = self.child.foreground_cmdline
+        conn_data = get_connection_data(args)
+        if conn_data is None:
+            get_boss().show_error('Could not handle remote file', 'No SSH connection data found in: {args}')
+            return
+        get_boss().run_kitten(
+            'remote_file', '--hostname', netloc.partition(':')[0], '--path', remote_path,
+            '--ssh-connection-data', json.dumps(conn_data)
+        )
+
     def focus_changed(self, focused: bool) -> None:
         if self.destroyed:
             return
+        call_watchers(weakref.ref(self), 'on_focus_change', {'focused': focused})
+        self.screen.focus_changed(focused)
         if focused:
+            changed = self.needs_attention
             self.needs_attention = False
-            if self.screen.focus_tracking_enabled:
-                self.screen.send_escape_code_to_child(CSI, 'I')
-        else:
-            if self.screen.focus_tracking_enabled:
-                self.screen.send_escape_code_to_child(CSI, 'O')
+            if changed:
+                tab = self.tabref()
+                if tab is not None:
+                    tab.relayout_borders()
 
     def title_changed(self, new_title: Optional[str]) -> None:
         self.child_title = sanitize_title(new_title or self.default_title)
@@ -461,27 +658,42 @@ class Window:
     def is_active(self) -> bool:
         return get_boss().active_window is self
 
+    @property
+    def has_activity_since_last_focus(self) -> bool:
+        return self.screen.has_activity_since_last_focus()
+
+    def on_activity_since_last_focus(self) -> None:
+        if get_options().tab_activity_symbol:
+            get_boss().on_activity_since_last_focus(self)
+
     def on_bell(self) -> None:
-        if self.opts.command_on_bell and self.opts.command_on_bell != ['none']:
-            import subprocess
+        cb = get_options().command_on_bell
+        if cb and cb != ['none']:
             import shlex
+            import subprocess
             env = self.child.final_env
             env['KITTY_CHILD_CMDLINE'] = ' '.join(map(shlex.quote, self.child.cmdline))
-            subprocess.Popen(self.opts.command_on_bell, env=env, cwd=self.child.foreground_cwd)
+            subprocess.Popen(cb, env=env, cwd=self.child.foreground_cwd)
         if not self.is_active:
+            changed = not self.needs_attention
             self.needs_attention = True
             tab = self.tabref()
             if tab is not None:
+                if changed:
+                    tab.relayout_borders()
                 tab.on_bell(self)
 
     def change_titlebar_color(self) -> None:
-        val = self.opts.macos_titlebar_color
+        opts = get_options()
+        val = opts.macos_titlebar_color if is_macos else opts.wayland_titlebar_color
         if val:
             if (val & 0xff) == 1:
                 val = self.screen.color_profile.default_bg
             else:
                 val = val >> 8
             set_titlebar_color(self.os_window_id, val)
+        else:
+            set_titlebar_color(self.os_window_id, 0, True)
 
     def change_colors(self, changes: Dict[DynamicColor, Optional[str]]) -> None:
         dirtied = default_bg_changed = False
@@ -511,6 +723,9 @@ class Window:
         b |= b << 8
         self.screen.send_escape_code_to_child(OSC, '{};rgb:{:04x}/{:04x}/{:04x}'.format(code, r, g, b))
 
+    def report_notification_activated(self, identifier: str) -> None:
+        self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
+
     def set_dynamic_color(self, code: int, value: Union[str, bytes]) -> None:
         if isinstance(value, bytes):
             value = value.decode('utf-8')
@@ -554,7 +769,8 @@ class Window:
             self.refresh()
 
     def request_capabilities(self, q: str) -> None:
-        self.screen.send_escape_code_to_child(DCS, get_capabilities(q))
+        for result in get_capabilities(q, get_options()):
+            self.screen.send_escape_code_to_child(DCS, result)
 
     def handle_remote_cmd(self, cmd: str) -> None:
         get_boss().handle_remote_cmd(cmd, self)
@@ -568,17 +784,37 @@ class Window:
     def send_cmd_response(self, response: Any) -> None:
         self.screen.send_escape_code_to_child(DCS, '@kitty-cmd' + json.dumps(response))
 
-    def clipboard_control(self, data: str) -> None:
+    def clipboard_control(self, data: str, is_partial: bool = False) -> None:
         where, text = data.partition(';')[::2]
+        if is_partial:
+            if self.clipboard_pending is None:
+                self.clipboard_pending = ClipboardPending(where, text)
+            else:
+                self.clipboard_pending = self.clipboard_pending._replace(data=self.clipboard_pending[1] + text)
+                if len(self.clipboard_pending.data) > 8 * 1024 * 1024:
+                    log_error('Discarding part of too large OSC 52 paste request')
+                    self.clipboard_pending = self.clipboard_pending._replace(data='', truncated=True)
+            return
+
         if not where:
-            where = 's0'
+            if self.clipboard_pending is not None:
+                text = self.clipboard_pending.data + text
+                where = self.clipboard_pending.where
+                try:
+                    if self.clipboard_pending.truncated:
+                        return
+                finally:
+                    self.clipboard_pending = None
+            else:
+                where = 's0'
+        cc = get_options().clipboard_control
         if text == '?':
             response = None
             if 's' in where or 'c' in where:
-                response = get_clipboard_string() if 'read-clipboard' in self.opts.clipboard_control else ''
+                response = get_clipboard_string() if 'read-clipboard' in cc else ''
                 loc = 'c'
             elif 'p' in where:
-                response = get_primary_selection() if 'read-primary' in self.opts.clipboard_control else ''
+                response = get_primary_selection() if 'read-primary' in cc else ''
                 loc = 'p'
             response = response or ''
             from base64 import standard_b64encode
@@ -592,25 +828,13 @@ class Window:
             except Exception:
                 text = ''
 
-            def write(key: str, func: Callable[[str], None]) -> None:
-                if text:
-                    if ('no-append' in self.opts.clipboard_control or
-                            len(self.clipboard_control_buffers[key]) > 1024*1024):
-                        self.clipboard_control_buffers[key] = ''
-                    self.clipboard_control_buffers[key] += text
-                else:
-                    self.clipboard_control_buffers[key] = ''
-                func(self.clipboard_control_buffers[key])
-
             if 's' in where or 'c' in where:
-                if 'write-clipboard' in self.opts.clipboard_control:
-                    write('c', set_clipboard_string)
+                if 'write-clipboard' in cc:
+                    set_clipboard_string(text)
             if 'p' in where:
-                if self.opts.copy_on_select == 'clipboard':
-                    if 'write-clipboard' in self.opts.clipboard_control:
-                        write('c', set_clipboard_string)
-                if 'write-primary' in self.opts.clipboard_control:
-                    write('p', set_primary_selection)
+                if 'write-primary' in cc:
+                    set_primary_selection(text)
+        self.clipboard_pending = None
 
     def manipulate_title_stack(self, pop: bool, title: str, icon: Any) -> None:
         if title:
@@ -623,10 +847,42 @@ class Window:
                     self.title_stack.append(self.child_title)
     # }}}
 
+    # mouse actions {{{
+    @ac('mouse', 'Click the URL under the mouse')
+    def mouse_click_url(self) -> None:
+        click_mouse_url(self.os_window_id, self.tab_id, self.id)
+
+    @ac('mouse', 'Click the URL under the mouse only if the screen has no selection')
+    def mouse_click_url_or_select(self) -> None:
+        if not self.screen.has_selection():
+            self.mouse_click_url()
+
+    @ac('mouse', '''
+        Manipulate the selection based on the current mouse position
+
+        For examples, see :ref:`conf-kitty-mouse.mousemap`
+        ''')
+    def mouse_selection(self, code: int) -> None:
+        mouse_selection(self.os_window_id, self.tab_id, self.id, code, self.current_mouse_event_button)
+
+    @ac('mouse', 'Paste the current primary selection')
+    def paste_selection(self) -> None:
+        txt = get_boss().current_primary_selection()
+        if txt:
+            self.paste(txt)
+
+    @ac('mouse', 'Paste the current primary selection or the clipboard if no selection is present')
+    def paste_selection_or_clipboard(self) -> None:
+        txt = get_boss().current_primary_selection_or_clipboard()
+        if txt:
+            self.paste(txt)
+    # }}}
+
     def text_for_selection(self) -> str:
         lines = self.screen.text_for_selection()
-        if self.opts.strip_trailing_spaces == 'always' or (
-                self.opts.strip_trailing_spaces == 'smart' and not self.screen.is_rectangle_select()):
+        sts = get_options().strip_trailing_spaces
+        if sts == 'always' or (
+                sts == 'smart' and not self.screen.is_rectangle_select()):
             return ''.join((ln.rstrip() or '\n') for ln in lines)
         return ''.join(lines)
 
@@ -652,29 +908,10 @@ class Window:
         as_ansi: bool = False,
         add_history: bool = False,
         add_wrap_markers: bool = False,
-        alternate_screen: bool = False
+        alternate_screen: bool = False,
+        add_cursor: bool = False
     ) -> str:
-        lines: List[str] = []
-        add_history = add_history and not (self.screen.is_using_alternate_linebuf() ^ alternate_screen)
-        if alternate_screen:
-            f = self.screen.as_text_alternate
-        else:
-            f = self.screen.as_text_non_visual if add_history else self.screen.as_text
-        f(lines.append, as_ansi, add_wrap_markers)
-        if add_history:
-            h: List[str] = []
-            self.screen.historybuf.pagerhist_as_text(h.append)
-            if h and (not as_ansi or not add_wrap_markers):
-                sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
-                h = list(map(sanitizer, h))
-            self.screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
-            if h:
-                if not self.screen.linebuf.is_continued(0):
-                    h[-1] += '\n'
-                if as_ansi:
-                    h[-1] += '\x1b[m'
-            return ''.join(chain(h, lines))
-        return ''.join(lines)
+        return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen, add_cursor)
 
     @property
     def cwd_of_child(self) -> Optional[str]:
@@ -698,30 +935,11 @@ class Window:
 
     # actions {{{
 
+    @ac('cp', 'Show scrollback in a pager like less')
     def show_scrollback(self) -> None:
         text = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
         data = self.pipe_data(text, has_wrap_markers=True)
-
-        def prepare_arg(x: str) -> str:
-            x = x.replace('INPUT_LINE_NUMBER', str(data['input_line_number']))
-            x = x.replace('CURSOR_LINE', str(data['cursor_y']))
-            x = x.replace('CURSOR_COLUMN', str(data['cursor_x']))
-            return x
-
-        cmd = list(map(prepare_arg, self.opts.scrollback_pager))
-        if not os.path.isabs(cmd[0]):
-            import shutil
-            exe = shutil.which(cmd[0])
-            if not exe:
-                env = read_shell_environment(self.opts)
-                if env and 'PATH' in env:
-                    exe = shutil.which(cmd[0], path=env['PATH'])
-                    if exe:
-                        cmd[0] = exe
-        bdata: Union[str, bytes, None] = data['text']
-        if isinstance(bdata, str):
-            bdata = bdata.encode('utf-8')
-        get_boss().display_scrollback(self, bdata, cmd)
+        get_boss().display_scrollback(self, data['text'], data['input_line_number'])
 
     def paste_bytes(self, text: Union[str, bytes]) -> None:
         # paste raw bytes without any processing
@@ -729,6 +947,7 @@ class Window:
             text = text.encode('utf-8')
         self.screen.paste_bytes(text)
 
+    @ac('cp', 'Paste the specified text into the current window')
     def paste(self, text: Union[str, bytes]) -> None:
         if text and not self.destroyed:
             if isinstance(text, str):
@@ -745,24 +964,34 @@ class Window:
                 text = text.replace(b'\r\n', b'\n').replace(b'\n', b'\r')
             self.screen.paste(text)
 
+    @ac('cp', 'Copy the selected text from the active window to the clipboard')
     def copy_to_clipboard(self) -> None:
         text = self.text_for_selection()
         if text:
             set_clipboard_string(text)
 
+    def encoded_key(self, key_event: KeyEvent) -> bytes:
+        return encode_key_for_tty(
+            key=key_event.key, shifted_key=key_event.shifted_key, alternate_key=key_event.alternate_key,
+            mods=key_event.mods, action=key_event.action, text=key_event.text,
+            key_encoding_flags=self.screen.current_key_encoding_flags(),
+            cursor_key_mode=self.screen.cursor_key_mode,
+        ).encode('ascii')
+
+    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection, send Ctrl-C')
     def copy_or_interrupt(self) -> None:
         text = self.text_for_selection()
         if text:
             set_clipboard_string(text)
         else:
-            mode = keyboard_mode_name(self.screen)
-            data = extended_key_event(defines.GLFW_KEY_C, defines.GLFW_MOD_CONTROL, defines.GLFW_PRESS) if mode == 'kitty' else b'\x03'
-            self.write_to_child(data)
+            self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
+    @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send Ctrl-C')
     def copy_and_clear_or_interrupt(self) -> None:
         self.copy_or_interrupt()
         self.screen.clear_selection()
 
+    @ac('cp', 'Pass the selected text from the active window to the specified program')
     def pass_selection_to_program(self, *args: str) -> None:
         cwd = self.cwd_of_child
         text = self.text_for_selection()
@@ -772,30 +1001,37 @@ class Window:
             else:
                 open_url(text, cwd=cwd)
 
+    @ac('sc', 'Scroll up by one line')
     def scroll_line_up(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_LINE, True)
 
+    @ac('sc', 'Scroll down by one line')
     def scroll_line_down(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_LINE, False)
 
+    @ac('sc', 'Scroll up by one page')
     def scroll_page_up(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_PAGE, True)
 
+    @ac('sc', 'Scroll down by one page')
     def scroll_page_down(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_PAGE, False)
 
+    @ac('sc', 'Scroll to the top of the scrollback buffer')
     def scroll_home(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_FULL, True)
 
+    @ac('sc', 'Scroll to the bottom of the scrollback buffer')
     def scroll_end(self) -> None:
         if self.screen.is_main_linebuf():
             self.screen.scroll(SCROLL_FULL, False)
 
+    @ac('mk', 'Toggle the current marker on/off')
     def toggle_marker(self, ftype: str, spec: Union[str, Tuple[Tuple[int, str], ...]], flags: int) -> None:
         from .marks import marker_from_spec
         key = ftype, spec
@@ -806,8 +1042,8 @@ class Window:
         self.current_marker_spec = key
 
     def set_marker(self, spec: Union[str, Sequence[str]]) -> None:
-        from .config import toggle_marker, parse_marker_spec
         from .marks import marker_from_spec
+        from .options.utils import parse_marker_spec, toggle_marker
         if isinstance(spec, str):
             func, (ftype, spec_, flags) = toggle_marker('toggle_marker', spec)
         else:
@@ -816,14 +1052,23 @@ class Window:
         self.screen.set_marker(marker_from_spec(ftype, spec_, flags))
         self.current_marker_spec = key
 
+    @ac('mk', 'Remove a previously created marker')
     def remove_marker(self) -> None:
         if self.current_marker_spec is not None:
             self.screen.set_marker()
             self.current_marker_spec = None
 
+    @ac('mk', 'Scroll to the next or previous mark of the specified type')
     def scroll_to_mark(self, prev: bool = True, mark: int = 0) -> None:
         self.screen.scroll_to_next_mark(mark, prev)
 
+    @ac('misc', '''
+        Send the specified SIGNAL to the foreground process in the active window
+
+        For example::
+
+            map F1 signal_child SIGTERM
+        ''')
     def signal_child(self, *signals: int) -> None:
         pid = self.child.pid_for_cwd
         if pid is not None:
